<?php

namespace Drupal\atoms;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Render\Renderer;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Utility\Token;
use Symfony\Component\DependencyInjection\ContainerInterface;

abstract class AtomsPluginBase extends PluginBase implements AtomsPluginInterface, ContainerFactoryPluginInterface
{

  use StringTranslationTrait;

  /** @var BubbleableMetadata */
  protected $bubbleable_metadata;

  protected $config;

  protected $moduleHandler;

  protected $token;

  public function __construct(ModuleHandlerInterface $moduleHandler, Token $token, ImmutableConfig $config, array $configuration, $plugin_id, $plugin_definition)
  {
    $this->config = $config;
    $this->moduleHandler = $moduleHandler;
    $this->token = $token;
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  static public function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition)
  {
    $configFactory = $container->get('config.factory');
    $config = $configFactory->get('atoms.settings');
    $moduleHandler = $container->get('module_handler');
    $token = $container->get('token');
    return new static($moduleHandler, $token, $config, $configuration, $plugin_id, $plugin_definition);
  }

  /**
   * @inheritDoc
   */
  public function validate(&$form, FormStateInterface $form_state)
  {
    return TRUE;
  }

  /**
   * @inheritDoc
   */
  public function submit(Atom $atom, FormStateInterface $form_state)
  {
    $atom->setData($form_state->getValue($this->getFormStateKey($atom)));
  }

  /**
   * @param \Drupal\atoms\Atom $atom
   *
   * @return array
   */
  protected function getFormStateKey(Atom $atom)
  {
    if (empty($atom->getSubgroupName())) {
      $subgroup = '0';
    } else {
      $subgroup = $atom->getSubgroupID();
    }
    return [$subgroup, $atom->getMachineName()];
  }

  /**
   * @inheritDoc
   */
  public function values(ViewableAtom $view)
  {
    return [
      'value' => $this->value($view)
    ];
  }

  /**
   * @inheritDoc
   */
  public function value(ViewableAtom $view, $key = '')
  {
    /** @var Renderer $renderer */
    static $renderer = NULL;
    if (is_null($renderer)) {
      $renderer = \Drupal::service('renderer');
    }
    $build = $this->render($view);
    return $renderer->renderPlain($build);
  }

  /**
   * @inheritDoc
   */
  public function render(ViewableAtom $view)
  {
    $this->bubbleable_metadata = new BubbleableMetadata();
    $build = [];
    $this->renderBuild($view, $build);
    $this->bubbleable_metadata->applyTo($build);
    return $build;
  }

  /**
   * @param \Drupal\atoms\ViewableAtom $view
   * @param array $build
   */
  abstract function renderBuild(ViewableAtom $view, array &$build);

  /**
   * @inheritDoc
   */
  public function getRenderSettings($setting)
  {
    $definitions = $this->getPluginDefinition();
    if (!empty($definitions['render_settings'][$setting])) {
      return $definitions['render_settings'][$setting];
    }
    return NULL;
  }

  /**
   * @inheritDoc
   */
  public function settingsForm(Config $config, FormStateInterface $form_state)
  {
    return [];
  }

  /**
   * @inheritDoc
   */
  public function settingsValidate(Config $config, FormStateInterface $form_state)
  {
    return TRUE;
  }

  /**
   * @inheritDoc
   */
  public function settingsSubmit(Config $config, FormStateInterface $form_state)
  {

  }

  /**
   * @param \Drupal\atoms\ViewableAtom $view
   * @param $text
   *
   * @return string
   */
  protected function prepareText(ViewableAtom $view, $text)
  {
    if ($this->isTokenSupportEnabled()) {
      return $this->token
        ->replace($text, $view->getTokenData(), [], $this->bubbleable_metadata);
    } else {
      return $text;
    }
  }

  /**
   * @return bool
   */
  protected function isTokenSupportEnabled()
  {
    if ($this->moduleHandler
        ->moduleExists('token') && $this->getConfig()->get('tokens')) {
      return TRUE;
    } else {
      return FALSE;
    }
  }

  /**
   * @return ImmutableConfig
   */
  protected function getConfig()
  {
    return $this->config;
  }

}
