<?php

namespace Drupal\atoms\Form;

use Drupal\atoms\Atom;
use Drupal\atoms\AtomsStorageException;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;

class AtomsEditForm extends FormBase
{

  /**
   * @inheritDoc
   */
  public function buildForm(array $form, FormStateInterface $form_state, $group_id = NULL, $langcode = NULL)
  {
    if (empty($group_id)) {
      return $form;
    }

    $atoms = Atom::loadGroup($group_id, $langcode);
    if (empty($atoms)) {
      return $form;
    } else {
      $form['group_id'] = [
        '#type' => 'hidden',
        '#value' => $group_id,
      ];
      $form['langcode'] = [
        '#type' => 'hidden',
        '#value' => $langcode,
      ];
    }
    $form['#tree'] = TRUE;

    $subgroups = ['0' => '0'];
    $token_types = [];
    $token_support = FALSE;
    foreach ($atoms as $id => $atom) {
      if ($this->getFormId() == 'atoms_translation_form' && !$atom->isTranslatable()) {
        unset($atoms[$id]);
        continue;
      }
      if (!empty($atom->getSubgroupName())) {
        $subgroups[$atom->getSubgroupID()] = $atom->getSubgroupName();
      }
      if ($atom->getPlugin()->hasTokenSupport()) {
        $token_support = TRUE;
        if (!empty($atom->getTokenTypes())) {
          $token_types = array_merge($token_types, $atom->getTokenTypes());
        }
      }
    }
    if (ksort($subgroups)) {
      foreach ($subgroups as $id => $subgroup) {
        if (empty($id)) {
          $form[$id] = [
            '#type' => 'container',
            '#tree' => TRUE
          ];
        } else {
          $form[$id] = [
            '#type' => 'fieldset',
            '#title' => $this->t($subgroup)
          ];
        }
      }
    }
    foreach ($atoms as $atom) {
      $type = $atom->getType();
      if (empty($atom->getSubgroupName())) {
        $subgroup = '0';
      } else {
        $subgroup = $atom->getSubgroupID();
      }

      if (!empty($langcode) && $atom->isTranslatable()) {
        $atom->setLangcode($langcode);
      }
      $atom->setPlugin(\Drupal::service('plugin.atoms.manager')->createInstanceFromType($type));
      $form[$subgroup][$atom->getMachineName()] = $atom->getPlugin()
        ->formBuilder($atom);
      $form[$subgroup][$atom->getMachineName()]['#weight'] = $atom->getWeight();
    }

    if (\Drupal::moduleHandler()
        ->moduleExists('token') && $this->config('atoms.settings')
        ->get('tokens') && $token_support) {
      $form['token_help'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => $token_types,
      ];
    }

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save changes'),
      '#button_type' => 'primary',
    ];

    return $form;
  }

  /**
   * @inheritDoc
   */
  public function getFormId()
  {
    return 'atoms_edit_form';
  }

  /**
   * @inheritDoc
   */
  public function validateForm(array &$form, FormStateInterface $form_state)
  {
    $result = TRUE;
    $atoms = $this->getGroup($form_state);
    /** @var Atom $atom */
    foreach ($atoms as $atom) {
      $result &= $atom->getPlugin()->validate($form, $form_state);
    }
  }

  protected function getGroup(FormStateInterface $formState)
  {
    $group_id = $formState->getValue('group_id');
    $langcode = empty($formState->getValue('langcode')) ? NULL : $formState->getValue('langcode');
    $atoms = Atom::loadGroup($group_id, $langcode);
    /** @var Atom $atom */
    foreach ($atoms as $atom) {
      if (!empty($langcode)) {
        $atom->setLangcode($langcode);
      }
    }
    return $atoms;
  }

  /**
   * @inheritDoc
   */
  public function submitForm(array &$form, FormStateInterface $form_state)
  {
    if (isset($form_state->getUserInput()['op'])) {
      $success = TRUE;
      $category = '';
      $group = '';
      $atoms = $this->getGroup($form_state);
      /** @var Atom $atom */
      foreach ($atoms as $atom) {
        $atom->getPlugin()->submit($atom, $form_state);
        $category = $atom->getCategory();
        $group = $atom->getGroupName();
        try {
          if (!$atom->save()) {
            $success = FALSE;
          }
        } catch (AtomsStorageException $e) {
          $success = FALSE;
        }
        Cache::invalidateTags(['atoms:' . $atom->getMachineName()]);
      }
      if ($success) {
        \Drupal::messenger()
          ->addStatus($this->t('@category: %group saved!', [
            '@category' => $category,
            '%group' => $group,
          ]));
      } else {
        \Drupal::messenger()
          ->addError($this->t('An error occured saving @category: %group', [
            '@category' => $category,
            '%group' => $group,
          ]));
      }
      $form_state->setRedirect('atoms.overview');
    }
  }

  public function getTitle($group_id, $langcode = NULL)
  {
    $group_name = \Drupal::service('atoms.storage')->getGroupName($group_id);
    return $this->t('Edit %label', [
      '%label' => $this->t($group_name)->__toString(),
    ],
      ['langcode' => $langcode]);
  }

}
