<?php

namespace Drupal\atoms\Form;

use Drupal\atoms\Atom;
use Drupal\atoms\AtomsManager;
use Drupal\atoms\AtomsStorage;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

abstract class AtomsOverviewBaseForm extends FormBase
{

  protected $storage;

  protected $manager;

  protected $requestStack;

  protected $atoms;

  public function __construct(AtomsStorage $storage, AtomsManager $manager, RequestStack $requestStack)
  {
    $this->storage = $storage;
    $this->manager = $manager;
    $this->requestStack = $requestStack;
  }

  public static function create(ContainerInterface $container)
  {
    return new static(
      $container->get('atoms.storage'),
      $container->get('plugin.atoms.manager'),
      $container->get('request_stack')
    );
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  public function buildForm(array $form, FormStateInterface $form_state)
  {
    $langcode = \Drupal::languageManager()->getCurrentLanguage()->getId();
    $category = $this->requestStack->getCurrentRequest()->query->get('category', '');
    $type = $this->requestStack->getCurrentRequest()->query->get('type', '');
    $this->atoms = $this->storage->searchAtoms([
      'langcode' => $langcode,
      'category' => $category,
      'type' => $type,
    ], $this->searchLimit());

    $form['filters'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['form--inline', 'clearfix'],
      ],
    ];

    $data = $this->storage->getCategories();
    $categories = ['' => $this->t('- Any -')];
    foreach ($data as $row) {
      $categories[$row] = $this->t($row);
    }
    asort($categories);
    $form['filters']['categories'] = [
      '#type' => 'select',
      '#options' => $categories,
      '#title' => $this->t('Category'),
      '#default_value' => $category,
    ];

    $data = $this->manager->getTypeNames();
    $types = ['' => $this->t('- Any -')];
    foreach ($data as $id => $row) {
      $types[$id] = $this->t($row);
    }
    asort($types);
    $form['filters']['types'] = [
      '#type' => 'select',
      '#options' => $types,
      '#title' => $this->t('Field types'),
      '#default_value' => $type,
    ];

    $form['filters']['actions'] = [
      '#type' => 'actions',
    ];
    $form['filters']['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Filter'),
    ];

    return $form;
  }

  /**
   * @return int
   */
  abstract protected function searchLimit();

  /**
   * Form submission handler.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitForm(array &$form, FormStateInterface $form_state)
  {
    $query = [];
    if (!empty($form_state->getValue('categories')) && $form_state->getValue('categories') != 'all') {
      $query['category'] = $form_state->getValue('categories');
    }
    if (!empty($form_state->getValue('types')) && $form_state->getValue('types') != 'all') {
      $query['type'] = $form_state->getValue('types');
    }
    $form_state->setRedirect($this->formRoute(), [], ['query' => $query]);
  }

  /**
   * @return string
   */
  abstract protected function formRoute();

  /**
   * @param \Drupal\atoms\Atom $atom
   *
   * @return array
   */
  protected function getOperations(Atom $atom)
  {
    $links = [];
    $links['edit'] = [
      'title' => $this->t('Edit'),
      'url' => Url::fromRoute('atoms.edit', ['group_id' => $atom->getGroupID()]),
    ];
    $links['translate'] = [
      'title' => $this->t('Translate'),
      'url' => Url::fromRoute('atoms.translate.overview', ['group_id' => $atom->getGroupID()]),
    ];
    return $links;
  }

}
