<?php

namespace Drupal\atoms;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Render\RenderableInterface;
use Drupal\Core\Render\Renderer;

class ViewableAtom implements RenderableInterface, CacheableDependencyInterface
{

  use DependencySerializationTrait;

  protected $machine_name = '';

  protected $langcode = LanguageInterface::LANGCODE_NOT_SPECIFIED;

  /** @var AtomsPluginInterface */
  protected $plugin;

  protected $type = '';

  /** @var Atom */
  protected $atom;

  protected $rendered;

  protected $options = [];

  protected $token_data = [];

  /** @var CacheableMetadata */
  protected $cacheabilityMetadata;

  protected $inline = FALSE;

  /**
   * ViewableAtom constructor.
   *
   * @param string $machine_name
   */
  public function __construct($machine_name, $langcode = NULL)
  {
    $this->langcode = $langcode;
    $this->atom = Atom::load($machine_name, $langcode);
    if (!empty($this->atom)) {
      $this->type = $this->atom->getType();
      $this->options = $this->atom->getOptions();
      $this->plugin = $this->atom->getPlugin();
      if (!empty($this->plugin)) {
        if ($this->plugin->getRenderSettings('inline')) {
          $this->inline = TRUE;
        }
      }
    }
  }

  /**
   * @return string
   */
  public function getType()
  {
    return $this->type;
  }

  /**
   * @return array
   */
  public function getOptions()
  {
    if (empty($this->getAtom())) {
      return [];
    }
    return $this->getAtom()->getOptions();
  }

  /**
   * @param array $options
   *
   * @return self
   */
  public function setOptions(array $options)
  {
    $this->getAtom()->setOptions($options);
    $this->rendered = NULL;
    return $this;
  }

  public function getAtom()
  {
    if (empty($this->atom)) {
      $this->atom = Atom::load($this->machine_name, $this->langcode);
    }
    return $this->atom;
  }

  /**
   * @return mixed
   */
  public function getData()
  {
    if (empty($this->getAtom())) {
      return [];
    }
    return $this->getAtom()->getData();
  }

  /**
   * @return string
   */
  public function getLangCode()
  {
    if (empty($this->langcode)) {
      $this->langcode = \Drupal::languageManager()->getCurrentLanguage(LanguageInterface::TYPE_CONTENT)->getId();
    }
    return $this->langcode;
  }

  /**
   * @param string $key
   * @param mixed $value
   *
   * @return self
   */
  public function setOption($key, $value)
  {
    $options = $this->getAtom()->getOptions();
    $options[$key] = $value;
    return $this->setOptions($options);
  }

  /**
   * @return array
   */
  public function getTokenData()
  {
    return $this->token_data;
  }

  /**
   * @param array $data
   *
   * @return self
   */
  public function setTokenData(array $data)
  {
    $this->token_data = $data;
    $this->rendered = NULL;
    return $this;
  }

  /**
   * @return array
   * @deprecated Use toRenderable() or toRenderableInline() instead
   */
  public function asRenderable()
  {
    if ($this->inline) {
      return $this->toRenderableInline();
    } else {
      return $this->toRenderable();
    }
  }

  /**
   * Returns a render array representation of the object, but inline.
   *
   * @return array
   *   A render array.
   */
  public function toRenderableInline()
  {
    $build = $this->getRendered();
    $this->getCacheableMetadata()->applyTo($build);
    return $build;
  }

  /**
   * @return array
   *   A render array.
   */
  private function getRendered()
  {
    if (empty($this->rendered)) {
      if (!empty($this->plugin)) {
        $this->rendered = $this->plugin->render($this);
      }
      if (empty($this->rendered)) {
        $this->rendered = $this->toEmptyRenderable();
      }
    }
    return $this->rendered;
  }

  private function toEmptyRenderable()
  {
    return ['#plain_text' => $this->machine_name];
  }

  /**
   * @return CacheableMetadata
   */
  public function getCacheableMetadata()
  {
    if (empty($this->cacheabilityMetadata)) {
      $this->cacheabilityMetadata = CacheableMetadata::createFromRenderArray($this->getRendered());
      $this->cacheabilityMetadata->addCacheableDependency($this->getAtom());
    }
    return $this->cacheabilityMetadata;
  }

  /**
   * Returns a render array representation of the object.
   *
   * @return array
   *   A render array.
   */
  public function toRenderable()
  {
    $build = [
      '#theme' => 'atom',
      '#atom' => $this->getAtom(),
    ];
    $build[] = $this->getRendered();
    $this->getCacheableMetadata()->applyTo($build);
    $build['#theme_wrappers'] = ['atom_element'];
    if (!empty($this->getAtom())) {
      $build['#contextual_links'] = [
        'atoms' => [
          'route_parameters' => ['group_id' => $this->getAtom()->getGroupID()],
        ],
      ];
    }
    return $build;
  }

  /**
   * @return string
   * @deprecated Use __toString() instead
   */
  public function asString()
  {
    return $this->__toString();
  }

  /**
   * @return string
   */
  public function __toString()
  {
    /** @var Renderer $renderer */
    static $renderer = NULL;
    if (is_null($renderer)) {
      $renderer = \Drupal::service('renderer');
    }
    $rendered = $this->getRendered();
    $string = $renderer->renderPlain($rendered);
    if (is_string($string)) {
      return $string;
    } else {
      return $string->__toString();
    }
  }

  /**
   * @return string
   */
  public function toString()
  {
    return $this->__toString();
  }

  /**
   * @param string $key
   *
   * @return mixed
   */
  public function toValue($key = '')
  {
    return $this->getAtom()->getPlugin()->value($this, $key);
  }

  /**
   * @return array
   */
  public function getValues()
  {
    return $this->getAtom()->getPlugin()->values($this);
  }

  /**
   * The cache contexts associated with this object.
   *
   * These identify a specific variation/representation of the object.
   *
   * Cache contexts are tokens: placeholders that are converted to cache keys by
   * the @cache_contexts_manager service. The replacement value depends on the
   * request context (the current URL, language, and so on). They're converted
   * before storing an object in cache.
   *
   * @return string[]
   *   An array of cache context tokens, used to generate a cache ID.
   *
   * @see \Drupal\Core\Cache\Context\CacheContextsManager::convertTokensToKeys()
   */
  public function getCacheContexts()
  {
    return $this->getCacheableMetadata()->getCacheContexts();
  }

  /**
   * The cache tags associated with this object.
   *
   * When this object is modified, these cache tags will be invalidated.
   *
   * @return string[]
   *   A set of cache tags.
   */
  public function getCacheTags()
  {
    return $this->getCacheableMetadata()->getCacheTags();
  }

  /**
   * The maximum age for which this object may be cached.
   *
   * @return int
   *   The maximum time in seconds that this object may be cached.
   */
  public function getCacheMaxAge()
  {
    return $this->getCacheableMetadata()->getCacheMaxAge();
  }

}
